#ifndef STENCIL_MANAGER_HPP
#define STENCIL_MANAGER_HPP

#include <mesh/IBoundaryDescriptor.hpp>
#include <mesh/IConnectivity.hpp>
#include <mesh/StencilArray.hpp>
#include <mesh/StencilDescriptor.hpp>

#include <memory>
#include <set>
#include <unordered_map>
#include <vector>

class StencilManager
{
 public:
  using BoundaryDescriptorList = std::vector<std::shared_ptr<const IBoundaryDescriptor>>;

 private:
  StencilManager()  = default;
  ~StencilManager() = default;

  static StencilManager* m_instance;

  struct Key
  {
    size_t connectivity_id;
    StencilDescriptor stencil_descriptor;
    BoundaryDescriptorList symmetry_boundary_descriptor_list;

    PUGS_INLINE bool
    operator==(const Key& k) const
    {
      if ((connectivity_id != k.connectivity_id) or (stencil_descriptor != k.stencil_descriptor) or
          (symmetry_boundary_descriptor_list.size() != k.symmetry_boundary_descriptor_list.size())) {
        return false;
      }

      std::set<std::string> boundary_descriptor_set;
      for (auto&& p_boundary_descriptor : symmetry_boundary_descriptor_list) {
        const std::string name = stringify(*p_boundary_descriptor);
        boundary_descriptor_set.insert(name);
      }

      std::set<std::string> k_boundary_descriptor_set;
      for (auto&& p_boundary_descriptor : k.symmetry_boundary_descriptor_list) {
        const std::string name = stringify(*p_boundary_descriptor);
        k_boundary_descriptor_set.insert(name);
      }

      return boundary_descriptor_set == k_boundary_descriptor_set;
    }
  };

  struct HashKey
  {
    size_t
    operator()(const Key& key) const
    {
      // We do not use the symmetry boundary set by now
      return ((std::hash<decltype(Key::connectivity_id)>()(key.connectivity_id)) ^
              (std::hash<std::decay_t<decltype(Key::stencil_descriptor.numberOfLayers())>>()(
                 key.stencil_descriptor.numberOfLayers())
               << 1));
    }
  };

  template <ItemType source_item_type, ItemType target_item_type>
  using StoredStencilTMap =
    std::unordered_map<Key, std::shared_ptr<const StencilArray<source_item_type, target_item_type>>, HashKey>;

  StoredStencilTMap<ItemType::cell, ItemType::cell> m_stored_cell_to_cell_stencil_map;
  StoredStencilTMap<ItemType::node, ItemType::cell> m_stored_node_to_cell_stencil_map;

  template <ItemType source_item_type, ItemType target_item_type>
  const StencilArray<source_item_type, target_item_type>& _getStencilArray(
    const IConnectivity& connectivity,
    const StencilDescriptor& stencil_descriptor,
    const BoundaryDescriptorList& symmetry_boundary_descriptor_list,
    StoredStencilTMap<source_item_type, target_item_type>& stored_source_to_target_stencil_map);

 public:
  static void create();
  static void destroy();

  PUGS_INLINE
  static StencilManager&
  instance()
  {
    Assert(m_instance != nullptr, "StencilManager was not created!");
    return *m_instance;
  }

  void deleteConnectivity(const size_t connectivity_id);

  const CellToCellStencilArray& getCellToCellStencilArray(
    const IConnectivity& i_connectivity,
    const StencilDescriptor& stencil_descriptor,
    const BoundaryDescriptorList& symmetry_boundary_descriptor_list = {});

  const NodeToCellStencilArray& getNodeToCellStencilArray(
    const IConnectivity& i_connectivity,
    const StencilDescriptor& stencil_descriptor,
    const BoundaryDescriptorList& symmetry_boundary_descriptor_list = {});

  StencilManager(const StencilManager&) = delete;
  StencilManager(StencilManager&&)      = delete;
};

#endif   // STENCIL_MANAGER_HPP
